#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "model_type.h"
#include "model_res.h"
#include "vmf_nnm_inference_app.h"
#include "vmf_nnm_fifoq_manager.h"
#include "my_demo_customize_inf_multiple_models.h"

#define TY_MAX_BOX_NUM (50)
#define FACE_SCORE_THRESHOLD 0.8f

// for face detection result, should be in DDR
static struct yolo_result_s *fd_result = NULL;
static struct kp_landmark_result_t *lm_result = NULL;


static bool is_init = false;

static int inference_face_detection(my_kl630_mul_example_header_t *input_header,
                                    struct yolo_result_s *fd_result /* output */)
{
    /******* Prepare the configuration *******/

    VMF_NNM_INFERENCE_APP_CONFIG_T inf_config;

    // Set the initial value of config to 0, false and NULL
    memset(&inf_config, 0, sizeof(VMF_NNM_INFERENCE_APP_CONFIG_T));

    // image buffer address should be just after the header
    inf_config.num_image = 1;
    inf_config.image_list[0].image_buf = (void *)((uint32_t)input_header + sizeof(my_kl630_mul_example_header_t));
    inf_config.image_list[0].image_width = input_header->width;
    inf_config.image_list[0].image_height = input_header->height;
    inf_config.image_list[0].image_channel = 3;                             // assume RGB565
    inf_config.image_list[0].image_format = KP_IMAGE_FORMAT_RGB565;         // assume RGB565
    inf_config.image_list[0].image_resize = KP_RESIZE_ENABLE;               // enable resize
    inf_config.image_list[0].image_padding = KP_PADDING_CORNER;             // enable padding on corner
    inf_config.image_list[0].image_norm = KP_NORMALIZE_KNERON;              // this depends on model
    inf_config.model_id = MY_KL630_SIN_EXAMPLE_FACE_DETECT_MODEL_ID;                   // this depends on model //수정해야함
    // set up fd result output buffer for ncpu/npu
    inf_config.ncpu_result_buf = (void *)fd_result;
    //inf_config.ncpu_result_buf = (void *)_pd_result;


    printf("Number of images: %d\n", inf_config.num_image);
    printf("Image Buffer: %p\n", inf_config.image_list[0].image_buf);
    printf("Image Width: %d\n", inf_config.image_list[0].image_width);
    printf("Image Height: %d\n", inf_config.image_list[0].image_height);
    printf("Image Channel: %d\n", inf_config.image_list[0].image_channel);
    printf("Model ID: %d (Modify as needed)\n", inf_config.model_id);

    /******* Activate inferencing in NCPU *******/
    printf("출력 버퍼 주소: %p\n", (void *)inf_config.ncpu_result_buf);

    printf("Check_ERROR 2 \n");

    if (inf_config.ncpu_result_buf == NULL) {

        printf("오류: 할당 후 출력 버퍼가 NULL입니다!\n");

        printf("Check_ERROR 2.1 \n"); 

    }

    int result = VMF_NNM_Inference_App_Execute(&inf_config); //Error!

    printf("추론 실행 후: 결과: %d\n", result);


    

    return result;
}

static int inference_face_landmarks(my_kl630_mul_example_header_t *input_header,
                                    struct bounding_box_s *face_box,
                                    kp_landmark_result_t *lm_result /* output */)
{
    /******* Prepare the configuration *******/

    VMF_NNM_INFERENCE_APP_CONFIG_T inf_config;

    // Set the initial value of config to 0, false and NULL
    memset(&inf_config, 0, sizeof(VMF_NNM_INFERENCE_APP_CONFIG_T));

    int32_t left = (int32_t)(face_box->x1);
    int32_t top = (int32_t)(face_box->y1);
    int32_t right = (int32_t)(face_box->x2);
    int32_t bottom = (int32_t)(face_box->y2);

    // image buffer address should be just after the header
    inf_config.model_id = MY_KL630_SIN_EXAMPLE_FACE_LANDMARK_MODEL_ID;                              // this depends on model //dmlwhsgodigka
    inf_config.num_image = 1;
    inf_config.image_list[0].image_buf = (void *)((uint32_t)input_header + sizeof(my_kl630_mul_example_header_t));
    inf_config.image_list[0].image_width = input_header->width;
    inf_config.image_list[0].image_height = input_header->height;
    inf_config.image_list[0].image_channel = 3;                                     // assume RGB565
    inf_config.image_list[0].image_format = KP_IMAGE_FORMAT_RGB565;                 // assume RGB565
    inf_config.image_list[0].image_norm = KP_NORMALIZE_KNERON;                      // this depends on model
    inf_config.image_list[0].image_resize = KP_RESIZE_ENABLE;                       // enable resize
    inf_config.image_list[0].image_padding = KP_PADDING_CORNER;                     // enable padding on corner
    inf_config.image_list[0].enable_crop = true;                                    // enable crop image in ncpu/npu

    // set crop box
    inf_config.image_list[0].crop_area.crop_number = 0;
    inf_config.image_list[0].crop_area.x1 = left;
    inf_config.image_list[0].crop_area.y1 = top;
    inf_config.image_list[0].crop_area.width = right - left;
    inf_config.image_list[0].crop_area.height = bottom - top;

    // set up landmark result output buffer for ncpu/npu
    inf_config.ncpu_result_buf = (void *)lm_result;

    /******* Activate inferencing in NCPU *******/

    return VMF_NNM_Inference_App_Execute(&inf_config);
}

static bool init_temp_buffer()
{
    // allocate DDR memory for ncpu/npu output restult
    //fd_result = (struct yolo_result_s *)kmdw_ddr_reserve(sizeof(struct yolo_result_s) + TY_MAX_BOX_NUM * sizeof(struct bounding_box_s));
    fd_result = (struct yolo_result_s *)calloc(1, sizeof(struct yolo_result_s));

    if (fd_result == NULL) {
        return false;
    }

    return true;
}

void my_kl630_mul_example_inf(int job_id, int num_input_buf, void **inf_input_buf_list)
{
    if (1 != num_input_buf) {
        VMF_NNM_Fifoq_Manager_Status_Code_Enqueue(job_id, KP_FW_WRONG_INPUT_BUFFER_COUNT_110);
        return;
    }


    int inf_status;
    int result_buf_size;
    uint32_t inf_result_buf;
    uint32_t inf_result_phy_addr;
    
    /******* Prepare the memory space of result *******/

    if (0 != VMF_NNM_Fifoq_Manager_Result_Get_Free_Buffer(&inf_result_buf, &inf_result_phy_addr, &result_buf_size, -1)) {
        printf("[%s] get result free buffer failed\n", __FUNCTION__);
        return;
    }


    printf("my kl630 code, start!\n");
    my_kl630_mul_example_header_t *input_header = (my_kl630_mul_example_header_t *)inf_input_buf_list[0];

    /******* Prepare the memory space of result *******/

    //void *inf_result_buf = kmdw_fifoq_manager_result_get_free_buffer(&result_buf_size);
    my_kl630_mul_example_result_t *output_result = (my_kl630_mul_example_result_t *)inf_result_buf;

    /******* Prepare header of output result *******/

    output_result->header_stamp.magic_type = KDP2_MAGIC_TYPE_INFERENCE;
    output_result->header_stamp.total_size = sizeof(my_kl630_mul_example_result_t);
    output_result->header_stamp.job_id = MY_KL630_MUL_EXAMPLE_JOB_ID;

    /******* Prepare the temporary memory space for the result of middle model *******/
    	
//    static bool is_init = false;

    if (!is_init) {
        int status = init_temp_buffer();
        if (!status) {
            // notify host error !
            output_result->header_stamp.status_code = KP_FW_DDR_MALLOC_FAILED_102;
            //kmdw_fifoq_manager_result_enqueue((void *)output_result, result_buf_size, false);
            printf("Error buf init error!");
            VMF_NNM_Fifoq_Manager_Result_Enqueue(inf_result_buf, inf_result_phy_addr, result_buf_size, -1, false);
            
            return;
        }

        is_init = true;
    }

    /******* Run face detect model *******/
    printf("[DEBUG] start inference previous");
    inf_status = inference_face_detection(input_header, fd_result);
    printf("success face detection inf!");
    if (inf_status != KP_SUCCESS) {
        // notify host error !
        printf("추론이 실패했습니다(결과 코드: %d\n)", inf_status);

        output_result->header_stamp.status_code = inf_status;
        printf("face detection failed!");
        VMF_NNM_Fifoq_Manager_Result_Enqueue(inf_result_buf, inf_result_phy_addr, result_buf_size, -1, false);
        return;
    }
    printf("face Detection SUCCESS\n");
    int face_cnt = 0;
    int max_face = (fd_result->box_count > FD_MAX) ? FD_MAX : fd_result->box_count;

    /******* Run landmark model for every faces *******/

    for (int i = 0; i < max_face; i++) {
        struct bounding_box_s *face_box = &fd_result->boxes[i];
        kp_landmark_result_t *face_lm_result = &output_result->faces[face_cnt].lm;

        if (FACE_SCORE_THRESHOLD < face_box->score) {
            // do face landmark for each faces
            inf_status = inference_face_landmarks(input_header, face_box, face_lm_result);

            if (KP_SUCCESS != inf_status) {
                // notify host error !
                output_result->header_stamp.status_code = inf_status;
                VMF_NNM_Fifoq_Manager_Result_Enqueue(inf_result_buf, inf_result_phy_addr, result_buf_size, -1, false);
                return;
            }

            // skip it if face lm is not good
            if (0.99f > face_lm_result->score) {
                continue;
            }

            memcpy(&output_result->faces[face_cnt].fd, face_box, sizeof(kp_bounding_box_t));
            face_cnt++;
        }
    }

    /******* Send the result to PLUS *******/
    printf("face landmark SUCCESS!");
    output_result->face_count = face_cnt;
    output_result->header_stamp.status_code = KP_SUCCESS;

    // send output result buffer back to host SW
    VMF_NNM_Fifoq_Manager_Result_Enqueue(inf_result_buf, inf_result_phy_addr, result_buf_size, -1, false);

}
static bool deinit_temp_buffer()
{
    if(is_init == true)
    {
        // free DDR memory for ncpu/npu output restult
        if(fd_result)
            free(fd_result);

        if(lm_result)
            free(lm_result);

        is_init = false;
    }

    return true;
}

void my_kl630_mul_example_inf_deinit()
{
    deinit_temp_buffer();
}
